# Generated by Django 3.1 on 2020-12-03 08:29

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("xlsx_import_tools", "0002_auto_20201130_1512"),
    ]

    operations = [
        migrations.AlterField(
            model_name="columndefinition",
            name="col_name",
            field=models.SlugField(
                help_text="A nicely formatted handle to refer to this column",
                unique=True,
            ),
        ),
    ]
