# Generated by Django 3.1 on 2020-12-03 13:00

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("xlsx_import_tools", "0003_auto_20201203_0829"),
    ]

    operations = [
        migrations.AlterField(
            model_name="sheetdefinition",
            name="sheet_name",
            field=models.TextField(
                help_text="A nicely formatted handle to refer to this sheet"
            ),
        ),
    ]
